@@ -22,7 +22,7 @@
 
 /**
  * JsonrpcClient:
- * 
+ *
  * A client for JSON-RPC communication
  *
  * The #JsonrpcClient class provides a convenient API to coordinate with a
@@ -240,7 +240,10 @@ error_invocations_from_idle (gpointer data)
 
   g_hash_table_iter_init (&iter, pd->invocations);
   while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&task))
-    g_task_return_error (task, g_error_copy (pd->error));
+    {
+      if (!g_task_get_completed (task))
+        g_task_return_error (task, g_error_copy (pd->error));
+    }
 
   g_clear_pointer (&pd->invocations, g_hash_table_unref);
   g_clear_pointer (&pd->error, g_error_free);
@@ -820,11 +823,20 @@ jsonrpc_client_call_read_cb (GObject      *object,
     {
       g_autoptr(GVariant) error_variant = NULL;
       g_autofree gchar *errstr = NULL;
+      const char *errmsg = NULL;
       gint64 id = -1;
+      gint64 errcode = -1;
 
       error_variant = g_variant_dict_lookup_value (dict, "error", NULL);
-      errstr = g_variant_print (error_variant, FALSE);
-      error = g_error_new_literal (G_IO_ERROR, G_IO_ERROR_FAILED, errstr);
+
+      if (error_variant != NULL &&
+          g_variant_lookup (error_variant, "message", "&s", &errmsg) &&
+          g_variant_lookup (error_variant, "code", "x", &errcode))
+        errstr = g_strdup_printf ("%s (%d)", errmsg, (int)errcode);
+      else
+        errstr = g_variant_print (error_variant, FALSE);
+
+      error = g_error_new_literal (JSONRPC_CLIENT_ERROR, errcode, errstr);
 
       if (g_variant_dict_lookup (dict, "id", "x", &id))
         {
@@ -1020,7 +1032,8 @@ jsonrpc_client_call_with_id_async (JsonrpcClient       *self,
   g_variant_dict_insert (&dict, "id", "x", idval);
   g_variant_dict_insert (&dict, "method", "s", method);
   g_variant_dict_insert_value (&dict, "params", params);
-  message = g_variant_dict_end (&dict);
+
+  message = g_variant_take_ref (g_variant_dict_end (&dict));
 
   g_hash_table_insert (priv->invocations, GINT_TO_POINTER (idval), g_object_ref (task));
 
@@ -1134,7 +1147,7 @@ jsonrpc_client_send_notification_write_cb (GObject      *object,
  * @cancellable: (nullable): A #GCancellable or %NULL
  *
  * Synchronously calls @method with @params on the remote peer.
- * 
+ *
  * This function will not wait or expect a reply from the peer.
  *
  * If @params is floating then the reference is consumed.
@@ -1171,7 +1184,8 @@ jsonrpc_client_send_notification (JsonrpcClient  *self,
   g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
   g_variant_dict_insert (&dict, "method", "s", method);
   g_variant_dict_insert_value (&dict, "params", params);
-  message = g_variant_dict_end (&dict);
+
+  message = g_variant_take_ref (g_variant_dict_end (&dict));
 
   ret = jsonrpc_output_stream_write_message (priv->output_stream, message, cancellable, error);
 
@@ -1186,7 +1200,7 @@ jsonrpc_client_send_notification (JsonrpcClient  *self,
  * @cancellable: (nullable): A #GCancellable or %NULL
  *
  * Asynchronously calls @method with @params on the remote peer.
- * 
+ *
  * This function will not wait or expect a reply from the peer.
  *
  * This function is useful when the caller wants to be notified that
@@ -1231,7 +1245,8 @@ jsonrpc_client_send_notification_async (JsonrpcClient       *self,
   g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
   g_variant_dict_insert (&dict, "method", "s", method);
   g_variant_dict_insert_value (&dict, "params", params);
-  message = g_variant_dict_end (&dict);
+
+  message = g_variant_take_ref (g_variant_dict_end (&dict));
 
   jsonrpc_output_stream_write_message_async (priv->output_stream,
                                              message,
@@ -1271,8 +1286,8 @@ jsonrpc_client_send_notification_finish (JsonrpcClient  *self,
  * @self: A #JsonrpcClient
  *
  * Closes the underlying streams and cancels any inflight operations of the
- * #JsonrpcClient. 
- * 
+ * #JsonrpcClient.
+ *
  * This is important to call when you are done with the
  * client so that any outstanding operations that have caused @self to
  * hold additional references are cancelled.
@@ -1519,7 +1534,8 @@ jsonrpc_client_reply (JsonrpcClient  *self,
   g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
   g_variant_dict_insert_value (&dict, "id", id);
   g_variant_dict_insert_value (&dict, "result", result);
-  message = g_variant_dict_end (&dict);
+
+  message = g_variant_take_ref (g_variant_dict_end (&dict));
 
   ret = jsonrpc_output_stream_write_message (priv->output_stream, message, cancellable, error);
 
@@ -1608,7 +1624,8 @@ jsonrpc_client_reply_async (JsonrpcClient       *self,
   g_variant_dict_insert (&dict, "jsonrpc", "s", "2.0");
   g_variant_dict_insert_value (&dict, "id", id);
   g_variant_dict_insert_value (&dict, "result", result);
-  message = g_variant_dict_end (&dict);
+
+  message = g_variant_take_ref (g_variant_dict_end (&dict));
 
   jsonrpc_output_stream_write_message_async (priv->output_stream,
                                              message,
